"""
Get Experiement run status from slurm outs
"""

import os, glob
import re
from datetime import datetime
import argparse
from .utils import box_str

parser = argparse.ArgumentParser(description='Arguemnts for plots')
parser.add_argument('-slurm_outs_folder', type=str, default='./slurm_outs', help='Slurm out folder path')
parser.add_argument('-box_width', type=int, default=100, help='width of the boxes')
parser.add_argument('-exp_name', type=str, default='all', help='Experiment name to search')
parser.add_argument('-save_path', type=str, default='', help='Save if specified.')


args = parser.parse_args()

files = glob.glob(args.slurm_outs_folder + '/*.out')

def title_str(title, box_width=None, indent=0):
    result = ''
    result += ' '*indent + f'{"─────────────────────────────────────────────────": ^{box_width}s}\n'
    result += ' '*indent + f'{title: ^{box_width}s}\n'
    result += ' '*indent + f'{"─────────────────────────────────────────────────": ^{box_width}s}\n'
    return result


#------------------ Parse slurm outs ------------------        
exp_status_dict = dict()

for fpath in files:
    with open(fpath, 'r') as f:
        sout = f.read()
        
        if not re.findall('Train: loss \d\.\d+ acc \d\.\d+; Val: loss \d\.\d+ acc \d\.\d+', sout): continue
        
        exp_infos = re.findall('Experiment : .*', sout)[0].split(' ')
        exp_name = exp_infos[2]
        
        if len(exp_infos)<4:
            i_splt, n_splt = 0, 1
        else:
            i_splt, n_splt = map(int, exp_infos[5][:-1].split('/'))
        
        if exp_name not in exp_status_dict:
            exp_status_dict[exp_name] = {
                'split_stats' : [{'alloc': 0, 
                                  'alive': 0, 
                                  'dead_list': [],
                                  'exp_counts': set(), 
                                  'tot_counts': 0,
                                  'elapsed_times': [],
                                  'last_epochs': [], 
                                  'n_epochs': 0,
                                  'train_accs': [],
                                  'val_accs': [], 
                                  } for _ in range(n_splt)],
                'total_stats' : {'alloc': 0,
                                 'alive': 0,
                                 'dead_list': [],
                                 'exp_counts': [set() for _ in range(n_splt)],
                                 'tot_counts': [*range(n_splt)],
                                },
            }
        split_dict = exp_status_dict[exp_name]['split_stats'][i_splt-1]
        tot_dict = exp_status_dict[exp_name]['total_stats']
        
        #------------------ current running job count ------------------
        ended = len(re.findall('##### END', sout))
        
        split_dict['alloc'] += 1
        split_dict['alive'] += 1-ended
        
        tot_dict['alloc'] += 1
        tot_dict['alive'] += 1-ended
        
        if ended:
            dead = os.path.split(fpath)[1][:-4]
            split_dict['dead_list'].append(dead)
            tot_dict['dead_list'].append(dead)
        
        #------------------ Experiment count statistics ------------------
        count_info = re.findall('Experiment count : \d+\/\d+', sout)
        get_counts = lambda s: (*map(int, s.split(' ')[3].split('/')),)
        
        split_dict['exp_counts'] |= {get_counts(s)[0]-1 for s in count_info}
        split_dict['tot_counts'] = get_counts(count_info[0])[1]
        
        tot_dict['exp_counts'][i_splt-1] |= {get_counts(s)[0]-1 for s in count_info}
        tot_dict['tot_counts'][i_splt-1] = get_counts(count_info[0])[1]

        #------------------ Elapsed time computation ------------------
        start_t, *_, end_t = re.findall('\d{4} \d\d:\d\d:\d\d', sout)
        s2t = lambda s: datetime.strptime(s, '%m%d %H:%M:%S')
        start_t, end_t = map(s2t, [start_t, end_t])
        elapsed_times = str(end_t - start_t)[:-3].replace(' day,', 'd')
        split_dict['elapsed_times'].append(elapsed_times)
        
        if 'start_t' not in tot_dict:
            tot_dict['start_t'] = start_t
            tot_dict['end_t'] = end_t
            
        tot_dict['start_t'] = min(start_t, tot_dict['start_t'])
        tot_dict['end_t'] = max(end_t, tot_dict['start_t'])
        
        #------------------ Latest experiment's epoch ------------------
        epoch_info = int(re.findall('Epoch \d+ \/ \d+', sout)[-1].split()[-3])
        n_epochs = int(re.findall('Epoch \d+ \/ \d+', sout)[-1].split()[-1])
        split_dict['last_epochs'].append(epoch_info)
        split_dict['n_epochs'] = n_epochs

        #------------------ Latest experiment's train / val accs ------------------
        metric_info = re.findall('Train: loss \d\.\d+ acc \d\.\d+; Val: loss \d\.\d+ acc \d\.\d+', sout)[-1]
        train_acc, val_acc = map(lambda x: str(float(x)*100)[:4]+'%', [metric_info.split()[4][:-1], metric_info.split()[-1]])
        split_dict['train_accs'].append(train_acc)
        split_dict['val_accs'].append(val_acc)
        
        exp_status_dict[exp_name]['split_stats'][i_splt-1] = split_dict
        exp_status_dict[exp_name]['total_stats'] = tot_dict

#------------------ Print stats ------------------       
 
summary = f'\nRun experiment status from Slurm outs (path : {args.slurm_outs_folder}):\n\n'

for i, (k, stats) in enumerate(exp_status_dict.items()):
    alive ,  alloc ,  dead_list ,  exp_counts ,  tot_counts ,  start_t ,  end_t = [stats['total_stats'][i] for i in \
  ['alive', 'alloc', 'dead_list', 'exp_counts', 'tot_counts', 'start_t', 'end_t']]
    
    exp_counts = sum(map(len, exp_counts))
    tot_counts = sum(tot_counts)

    summary += title_str(f"{i+1}. {k} summary", box_width=args.box_width, indent=5) + '\n'
    
    summary += box_str("",
                        f'''Experiment count status : {exp_counts} / {tot_counts} ({(p:=exp_counts/tot_counts)*100:.2f}%)
                            Alive / Allocated       : {alive} / {alloc} {f'(dead: {", ".join(dead_list)})' if dead_list else ''}
                            Max elapsed time        : {(t:=end_t-start_t)}
                            Expected time left      : {str(t*(1-p)/p)[:-7]}''',
                        boundaries='''   \n│  \n   ''',
                        box_width=args.box_width-9, indent=9, skip=1) + '\n'
    
    split_stats = stats['split_stats']
    for j, split_dict in enumerate(split_stats):
        alloc ,  alive ,  dead_list ,  exp_counts ,  tot_counts ,  elapsed_times ,  last_epochs ,  n_epochs ,  train_accs ,  val_accs  = [split_dict[i] for i in \
      ['alloc', 'alive', 'dead_list', 'exp_counts', 'tot_counts', 'elapsed_times', 'last_epochs', 'n_epochs', 'train_accs', 'val_accs']]
        
        exp_counts = sorted(exp_counts)
        all_exp_until = [*range(j, exp_counts[-1], len(split_stats))]
        skipover = [i for i in all_exp_until if i not in exp_counts]
        
        summary += box_str(f'Split ( {j+1} / {len(split_stats)} )', 
                            f'''Alive / Allocated : {alive} / {alloc} {f'(dead: {", ".join(dead_list)})' if dead_list else ''}
                                Experiments count status: {len(exp_counts)} / {tot_counts}
                                Elapsed times : {elapsed_times}
                                Skipped over experiments : {skipover} ({len(skipover)}/{len(all_exp_until)})
                                Current epochs : {last_epochs} (total epochs : {n_epochs})
                                Train accs : {train_accs}
                                Val accs : {val_accs}''',
                            box_width=args.box_width, indent=5) + '\n'
    summary += '\n'

if args.save_path:
    root, _ = os.path.split(args.save_path)
    if not os.path.exists(root):
        print(f'save_path does not exist : {root}')
    else:
        with open(args.save_path, 'w') as f:
            f.write(summary)
else:
    print(summary)
    


import os
import re
import yaml
from functools import partial
from itertools import product

from absl import app, flags
from ml_collections import ConfigDict

import matplotlib.pyplot as plt
import matplotlib.style as style
import seaborn as sns

from .experiment import Experiment, ExperimentLog
from .utils import str2value, df2richtable

from rich import print
from rich.panel import Panel
from rich.columns import Columns
from rich.align import Align

from .plot_utils.data_processor import avgbest_df, select_df
from .plot_utils.plot_drawer import ax_draw_curve, ax_draw_best_stared_curve, ax_draw_bar, ax_draw_heatmap
from .plot_utils.utils import merge_dict, default_style, ax_styler, save_figure

FLAGS = flags.FLAGS


def get_plot_config(plot_config: dict, plot_args: dict):
    assert plot_args['mode'] in plot_config, f'Mode: {plot_args["mode"]} does not exist.'
    alias_mode = ('-' not in plot_args['mode'])
    
    p_cfg = plot_config[plot_args['mode']]
    if alias_mode:
        p_cfg_base = plot_config.get(p_cfg['mode'], dict())
        p_cfg_base = merge_dict(p_cfg_base, plot_args)
        p_cfg_base = merge_dict(p_cfg_base, plot_config['default_style'])
        return merge_dict(p_cfg, p_cfg_base)
    else:
        return {**plot_args, **p_cfg}
    
def __save_name_builder(pflt, pmlf, pcfg, save_name=''):
    if pflt:
        save_name += pflt.replace(' / ', '-').replace(' ', '_')
    if pmlf:
        save_name = '-'.join([*pmlf, save_name])
    if any([pcfg[f'best_ref_{k}'] for k in ['x_fields', 'metric_field', 'ml_fields']]):
        save_name +=  f"-({pcfg['best_ref_x_fields']}, {pcfg['best_ref_metric_field']}, {pcfg['best_ref_ml_fields']})"
    save_name += "-max" if pcfg['best_at_max'] else "-min"
    return save_name


def draw_metric(tsv_file, plot_config, save_name='', preprcs_df=lambda *x: x):
        pcfg = plot_config
        
        # parse mode string
        mode, x_fields, metric = pcfg['mode'].split('-') # ex) {sam}-{epoch}-{train_loss}
        x_fields = x_fields.split(' ')
        
        pflt, pmlf = map(pcfg.get, ['filter', 'multi_line_fields'])
        
        # choose plot mode
        if mode=='curve':
            assert len(x_fields)==1, f'Number of x_fields shoud be 1 when using curve mode, but you passed {len(x_fields)}.'
            ax_draw = ax_draw_curve
            y_label = metric.replace('_', ' ').capitalize()
        elif mode=='curve_best':
            assert len(x_fields)==1, f'Number of x_fields shoud be 1 when using curve mode, but you passed {len(x_fields)}.'
            ax_draw = ax_draw_best_stared_curve
            y_label = metric.replace('_', ' ').capitalize()
        elif mode=='bar':
            assert len(x_fields)==1, f'Number of x_fields shoud be 1 when using bar mode, but you passed {len(x_fields)}.'
            ax_draw = ax_draw_bar
            y_label = metric.replace('_', ' ').capitalize()
        elif mode=='heatmap':
            assert len(x_fields)==2, f'Number of x_fields shoud be 2 when using heatmap mode, but you passed {len(x_fields)}.'
            assert not pmlf, f'No multi_line_fieldss are allowed in heatmap mode, but you passed {len(x_fields)}.'
            ax_draw = ax_draw_heatmap
            y_label = x_fields[1].replace('_', ' ').capitalize()
        
        # get dataframe, drop unused metrics for efficient process
        log = ExperimentLog.from_tsv(tsv_file)
        assert metric in log.df, f'Metric {metric} not in log. Choose between {list(log.df)}'
        
        #--- initial filter for df according to FLAGS.filter (except epoch and metric)
        if pflt:
            filt_dict = [*map(lambda flt: re.split('(?<!,) ', flt.strip()), pflt.split('/'))] # split ' ' except ', '
            log.df = select_df(log.df, {fk:[*map(str2value, fvs)] for fk, *fvs in filt_dict if fk not in {'step', 'metric'}})
        
        #--- melt and explode metric in log.df
        if 'metric' not in pmlf and 'metric' not in x_fields:
            log.df = log.df.drop(list(set(log.df)-{metric, pcfg['best_ref_metric_field']}), axis=1)
        df = log.melt_and_explode_metric(step=None if 'step' in x_fields or 'step' in pflt else -1)
        
        assert not df.empty, f'Metric {metric}' +\
            (f' and best_ref_metric_field {pcfg["best_ref_metric_field"]} are' if pcfg["best_ref_metric_field"] else ' is') +\
                ' NaN in given dataframe.'
        
        #---filter df according to FLAGS.filter step and metrics
        if pflt:
            e_rng = lambda fvs: [*range(*map(int, fvs[0].split(':')))] if (len(fvs)==1 and ':' in fvs[0]) else fvs # CNG 'a:b' step filter later
            df = select_df(df, {fk:[*map(str2value, e_rng(fvs))] for fk, *fvs in filt_dict if fk in {'step', 'metric'}}) 
        
        
        #---set mlines according to FLAGS.multi_line_fields
        if pmlf:
            mlines = product(*[sorted(set(df.index.get_level_values(f)), key=str2value) for f in pmlf])
        else:
            pmlf, mlines = ['metric'], [[metric]]
            pcfg['ax_style'].pop('legend', None)
        
        #---preprocess best_ref_x_fields
        pcfg['best_ref_x_fields'] = [*map(str2value, pcfg['best_ref_x_fields'])]
        
        best_over = set(df.index.names) - {*x_fields, 'metric', 'seed', *pmlf}
        best_at_max = pcfg['best_at_max']
        
        if 'step' in x_fields:
            if not pcfg['best_ref_x_fields']:
                pcfg['best_ref_x_fields'] = ['']*len(x_fields)
            i = x_fields.index('step')
            if 'step' in pflt:
                st = [i for i in filt_dict if 'step' in i][0][1]
                pcfg['best_ref_x_fields'][i] = int(st.split(':')[1])-1 if ':' in st else int(st) # CNG 'a:b' step filter later
            else:
                pcfg['best_ref_x_fields'][i]=min(*df.index.get_level_values('total_steps'))-1
                
        # build save name
        save_name = __save_name_builder(pflt, pmlf, pcfg, save_name=save_name)
        
        # Notify selected plot configs and field handling statistics
        specified_field = {k for k in best_over if len(set(df.index.get_level_values(k)))==1}
        print('\n\n',
            Align(
                Columns(
                    [Panel('\n'.join([f'- {k}: {pcfg[k]}' 
                                            for k in ('mode', 'multi_line_fields', 
                                                        'filter', 'best_at_max', 
                                                        'best_ref_x_fields', 'best_ref_metric_field', 
                                                        'best_ref_ml_fields') if pcfg[k]]),
                                    title='Plot configuration', padding=(1, 3)),
                            Panel(f"- Key field (has multiple values): {[*x_fields, *pmlf]} (2)\n" + \
                                  f"- Specified field: {(spf:=[*specified_field, 'metric'])} ({len(spf)})\n"+ \
                                  f"- Averaged field: {['seed']} (1)\n" + \
                                  f"- Optimized field: {(opf:=list(best_over-specified_field))} ({len(opf)})",
                                    title='Field handling statistics', padding=(1, 3))]
                        ), align='center'
                ))

        
        
        ############################# Prepare dataframe #############################
        
        best_of = {}
        if pcfg['best_ref_x_fields']: # same hyperparameter over all points in line
            best_of.update(dict([*zip(x_fields, pcfg['best_ref_x_fields'])]))

        if pcfg['best_ref_metric_field']: # Optimize in terms of reference metric, and apply those hyperparameters to original
            best_of['metric'] = pcfg['best_ref_metric_field']
        
        if pcfg['best_ref_ml_fields']: # same hyperparameter over all line in multi_line_fields
            best_of.update(dict([*zip(pmlf, pcfg['best_ref_ml_fields'])]))
            
        # change field name and avg over seed and get best result over best_over
        best_df = avgbest_df(df, 'metric_value',
                             avg_over='seed', 
                             best_over=best_over, 
                             best_of=best_of,
                             best_at_max=best_at_max)
        
        print('\n', Align(df2richtable(best_df), align='center'))
        
        if 'metric' not in pmlf and 'metric' not in x_fields:
            best_df = select_df(best_df, {'metric': metric})
        
        ############################# Plot #############################
        
        # prepare plot
        fig, ax = plt.subplots()
        
        if pcfg['colors']=='':
            colors = iter(sns.color_palette()*10)
        elif pcfg['colors']=='cont':
            colors = iter([c for i, c in enumerate(sum(map(sns.color_palette, ["light:#9467bd", "Blues", "rocket", "crest", "magma"]*3), [])[1:])])# if i%2])
        
        # select specified metric if multi_line_fields isn't metric
        if 'metric' not in pmlf:
            best_df = select_df(best_df, {'metric': metric}, *x_fields)
            
        for mlvs in mlines:
            try:
                p_df = select_df(best_df, {f: v for f, v in zip(pmlf, mlvs)}, *x_fields)
            except:
                continue
            legend = ','.join([(v if isinstance(v, str) else f'{f} {v}').replace('_', ' ') for f, v in zip(pmlf, mlvs)])
            
            p_df, legend, mlvs = preprcs_df(p_df, legend, mlvs)
            
            # remove unnessacery fields
            p_df = p_df.reset_index([*(set(p_df.index.names) - set(x_fields))], drop=False)
            if len(x_fields)>1:
                p_df = p_df.reorder_levels(x_fields)
            p_df = p_df.sort_index(key=lambda s: [*map(str2value, s)])
            p_df = p_df[['metric_value', *(set(p_df)-{'metric_value'})]]
            
            pcfg['line_style']['color'] = next(colors)
            ax = ax_draw(ax, p_df, 
                         label=legend,
                         annotate=pcfg['annotate'],
                         annotate_field=pcfg['annotate_field'],
                         std_plot=pcfg['std_plot'],
                         best_at_max=best_at_max,
                         **pcfg['line_style'])
        
        return fig, ax, y_label, save_name.strip('-')
    

def run(argv, preprcs_df):
    if len(argv)>2:
        raise app.UsageError('Too many command-line arguments.')
    
    # Preprocess plot_config
    flag_dict = FLAGS.flag_values_dict()
    
    plot_config = {**default_style, **flag_dict}
    if FLAGS.plot_config!='':
        with open(FLAGS.plot_config) as f:
            plot_config = yaml.safe_load(f.read())
            plot_config = get_plot_config(plot_config, flag_dict)
    
    # set ax_style related arguments
    fig_size = flag_dict.pop('fig_size')
    if fig_size:
        if len(fig_size)==1:
            fig_size = fig_size*2
        fig_size = [*map(float, fig_size)]
        plot_config['ax_style']['fig_size'] = fig_size
    
    title = flag_dict.pop('title')
    if title:
        plot_config['ax_style']['title'] = [title, {'size': flag_dict['font_size']}]
        
    xlabel = flag_dict.pop('xlabel')
    if xlabel:
        plot_config['ax_style']['xlabel'] = [xlabel, {'size': flag_dict['font_size']}]
        
    ylabel = flag_dict.pop('ylabel')
    if ylabel:
        plot_config['ax_style']['ylabel'] = [ylabel, {'size': flag_dict['font_size']}]
        
    
    # set style
    style.use(plot_config['style'])
    
    # get paths
    _, tsv_file, fig_dir = Experiment.get_paths(plot_config['exp_folder'])
    save_dir = os.path.join(fig_dir, plot_config['mode'])
    
    if plot_config['mode'].split('-')[0] in {'curve', 'curve_best', 'bar', 'heatmap'}:
        fig, ax, y_label, save_name = draw_metric(tsv_file, plot_config, preprcs_df=preprcs_df)
        ax.set_ylabel(y_label)
    else:
        assert False, f'Mode: {plot_config["mode"]} does not exist.'
    
    ax_styler(ax, **plot_config['ax_style'])
    save_figure(fig, save_dir, save_name)
    
    print('\n', Align(Panel(f'save plot at: {fig_dir}/[bold blue_violet]{plot_config["mode"]}[/bold blue_violet]/[bold spring_green1]{save_name}[/bold spring_green1].pdf', 
                      title='Plot complete', padding=(1, 3), expand=False), align='center'), '\n')

    
def main(preprcs_df = lambda *x: x):
    flags.DEFINE_string('exp_folder', '', "Experiment folder path.")
    
    flags.DEFINE_string('mode', 'curve-epoch-val_loss', "Plot mode.")
    flags.DEFINE_string('filter', '', "filter values.")
    flags.DEFINE_spaceseplist('multi_line_fields', '', "List of fields to plot multiple lines over.")
    flags.DEFINE_spaceseplist('best_ref_x_fields', '', "Reference x_field-values to evaluate optimal hyperparameters.")
    flags.DEFINE_string('best_ref_metric_field', '', "Reference metric_field-values to evaluate optimal hyperparameters.")
    flags.DEFINE_spaceseplist('best_ref_ml_fields', '', "Reference multi_line_fields-value to evaluate optimal hyperparameters.")
    flags.DEFINE_bool('best_at_max', False, 'Whether the bese metric value is the maximum value.')
    
    flags.DEFINE_string('plot_config', '', "Yaml file path for various plot setups.")
    flags.DEFINE_string('colors', '', "color scheme ('', 'cont').")
    flags.DEFINE_string('style', 'default', "Matplotlib style.")
    flags.DEFINE_bool('annotate', True, 'Run multiple plot according to given config.')
    flags.DEFINE_spaceseplist('annotate_field', '', 'List of fields to include in annotation.')
    flags.DEFINE_spaceseplist('fig_size', '', 'Figure size.')
    flags.DEFINE_string('title', '', "Title.")
    flags.DEFINE_string('xlabel', '', "Label of x-axis.")
    flags.DEFINE_string('ylabel', '', "Label of y-axis.")
    flags.DEFINE_integer('font_size', 22, "Font size of title and label.")

    app.run(partial(run, preprcs_df=preprcs_df))
    
    
if __name__=='__main__':
    main()



import pandas as pd

def select_df(df, filt_dict, *exclude_fields, drop=False):
    """Select df rows with matching from given filt_dict except ``exclude_fields``"""
    assert not df.empty, 'Given dataframe is empty.'
    assert not (k:=set(filt_dict.keys()) - set(df.index.names)), f'filt_dict keys {k} is not in df.'
    if not filt_dict: return df
    
    filt_keys = set(filt_dict.keys()) - set(exclude_fields)    # filter out exclude field
    
    nest = lambda vs: vs if isinstance(vs, list) else [vs]
    for i, k in enumerate(filt_keys):
        values = nest(filt_dict[k])
        assert not (v:=set(values)-(vs:=set(df.index.get_level_values(k)))), f"Values {v} are not in field '{k}': {sorted(vs)}"
        df = df.loc[df.index.get_level_values(k).isin(values)]
        assert not df.empty, f"Filter {k}:{values} return empty dataframe. Inspect {dict((k, filt_dict[k]) for k in filt_keys[:i+1])}" 
    
    if drop:
        df = df.reset_index([*filt_keys], drop=True)
        
    return df

def homogenize_df(df, ref_df, filt_dict, *exclude_fields):
    """Homogenize index values of ``df`` with reference to ``select_df(ref_df, filt_dict)``."""
    ref_idx = select_df(ref_df, filt_dict, drop=True).index
    slcted_dfs = [select_df(df, dict(zip(ref_idx.names, d))) for d in ref_idx.values]
    df = pd.concat(slcted_dfs)
    return df

def avgbest_df(df, metric_field,
               avg_over=None, 
               best_over=tuple(), 
               best_of=dict(), 
               best_at_max=True):
    """Average over ``avg_over`` and get best result over ``best_over``
    
    Args:
        df (pandas.DataFrame): Base dataframe to operate over. All hyperparameters should be set as `MultiIndex`.
        metric_field (str): Column name of the metric. Used to evaluate best hyperparameter.
        avg_over (str): `MultiIndex` level name to average over.
        best_over (List[str]): List of `MultiIndex` level names to find value yielding best values of `metric_field`.
        best_of (Dict[str, Any]): Dictionary of pair `{MultiIndex name}: {value in MultiIndex}` to find best hyperparameter of. The other values in `{MultiIndex name}` will follow the best hyperparamter found for these values.
        best_at_max (bool): `True` when larger metric is better, and `False` otherwise.
        
    Returns: 
        pandas.DataFrame: Processed DataFrame
    """
    '''
    - aggregate index : avg_over, best_over
    - key index : best_of, others
    '''
    df_fields = set(df.index.names)
    
    # avg over avg_over
    if avg_over is not None:
        df_fields -= {avg_over}
        avg_over_group = df.groupby([*df_fields], dropna=True)
        df = avg_over_group.mean(numeric_only=True)
        df[metric_field+'_std'] = avg_over_group.sem(numeric_only=True)[metric_field]  # add std column
    
    # best result over best_over
    if best_over:
        # find best result over best_over for best_of
        df_fields -= set(best_over)
        best_df = select_df(df, best_of)
        if df_fields:
            best_df = best_df.loc[best_df.groupby([*df_fields])[metric_field]
                                         .aggregate(('idxmin', 'idxmax')[best_at_max])]
        else: # need since groupby returns series and causes error when df_fields is empty
            best_df = best_df.loc[best_df[metric_field].aggregate(('idxmin', 'idxmax')[best_at_max])]
            

        # match best_over values for non-best_of-key-index with best_of-key-index
        df_fields -= set(best_of)
        df = homogenize_df(df, best_df, best_of, *df_fields)
    
    return df



import os, glob, shutil, io
import copy
import yaml
import re
from functools import reduce, partial
from typing import Optional, ClassVar, Callable, Union, TypeVar
from dataclasses import dataclass
from itertools import product, chain
from datetime import datetime
from git import Repo

import pandas as pd
import numpy as np
from rich.progress import track

from absl import logging
from ml_collections.config_dict import ConfigDict

from .utils import list2tuple, str2value

Self = TypeVar("Self", bound="Experiment")
ExpFunc = Union[Callable[[ConfigDict], dict], Callable[[ConfigDict, Self], dict]]

class ConfigIter:
  '''
  Iterator of configs generated from yaml config generator file
  Generates grid of configs from given experiment plans in yaml.
    
  [yaml exp_file example]
   model: ResNet32
   dataset: cifar10
   ...
   grid_fields: [optimizer, pai_type, pai_sparsity, rho, seed]
   grid:
     - optimizer: [sgd]
       group:
         pai_type: [[random], [snip]]
         pai_scope: [[local], [global]]
       rho: [0.05]
       seed: [1, 2, 3]
     - optimizer: [sam]
       pai_type: [random, snip, lth]
       pai_sparsity: [0, 0.9, 0.95, 0.98, 0.99]
       rho: [0.05, 0.1, 0.2, 0.3]
       seed : [1, 2, 3]
  '''
  
  def __init__(self, exp_config_path):
    with open(exp_config_path) as f:
      cnfg_str = self.__sub_cmd(f.read())
      self.grid_fields = self.__extract_grid_order(cnfg_str)
      self.raw_config = yaml.safe_load(cnfg_str)
    
    self.name = os.path.split(exp_config_path)[0].split('/')[-1]
    self.grid = self.raw_config.get('grid')
    self.static_configs = {k:self.raw_config[k] for k in set(self.raw_config)-{'grid_fields', 'grid'}}
  
    self.grid_iter = self.__get_iter()
    
  
  def filter_iter(self, filt_fn):
    """filters ConfigIter with ``filt_fn`` which has (idx, dict) as arguments"""
    self.grid_iter = [d for i, d in enumerate(self.grid_iter) if filt_fn(i, d)]
    
  @property
  def grid_dict(self):
    """dictionary of all grid values"""
    if not self.grid_fields: return dict()
    st, *sts = self.grid
    castl2t = lambda vs: map(lambda v: (tuple(v) if isinstance(v, list) else v), vs)
    acc = lambda a, d: {k: [*{*castl2t(va), *castl2t(vd)}] 
                        for (k, va), (_, vd) in zip(a.items(), d.items())}
    grid_dict = {k: [*map(self.field_type(k), vs)] for k, vs in reduce(acc, sts, st).items()}
    return grid_dict
  
  def field_type(self, field):
    """Returns type of a field"""
    return type(self[0][field])
  
  @staticmethod
  def __sub_cmd(cfg_str):
    '''compile special commands in experiment plan (study) of cfg_str'''
    
    # \[__;0:2:10--11:5:20] -> [__ for i in range(0, 10, 2)] + [__ for i in range(11, 20, 5)]
    for entry in re.finditer(p:='\[[^;\n]+;(\d+:\d+:\d+(--)?)+\]', cfg_str):
      f, rngs = entry.group()[1:-1].split(';')
      sdes = [map(int, rng.split(':')) for rng in rngs.split('--')]
      assert 'i' in f, f"Variable i should be in the expression '{entry}'"
      assert re.sub('[^\+\-\*/\[\]i\d\(\), ]', '', f)==f, f"Cannot use alphabet other than 'i' in expression '{entry}'"
      
      rep = sum([eval(f'[{f} for i in {range(s, e, d)}]') for s, d, e in sdes], start=[])
      cfg_str=re.sub(p, str(rep), cfg_str, 1)
    
    return cfg_str
    
  @staticmethod
  def __extract_grid_order(cfg_str):
    """parse grid order from raw config string"""
    grid = re.split('grid ?:', cfg_str)[1]
    names = re.findall('[\w_]+(?= ?:)', grid)
    
    dupless_names = []
    for n in names:
      if n in dupless_names or n=='group': continue
      dupless_names.append(n)
      
    return dupless_names
    
  @staticmethod
  def __ravel_group(study):
    # Return list of study if there is no 'group'
    if 'group' not in study: return [study]
    
    # Ravel grouped fields into list of experiment plans.
    grid_g = lambda g: ([*zip(g.keys(), value)] for value in zip(*g.values()))
    if type(study['group'])==dict:
      r_groups = grid_g(study['group'])
    elif type(study['group'])==list:
      r_groups = (chain(*gs) for gs in product(*map(grid_g, study['group'])))
    study.pop('group')
    
    raveled_study = ({**study, **dict(g)} for g in r_groups)
    return raveled_study

  def __get_iter(self):
    if self.grid_fields is None: return [dict()]
    
    # Prepare Experiment, create experiment plan (grid)
    if type(self.grid)==dict:
      self.grid = [*self.__ravel_group(self.grid)]
    elif type(self.grid)==list:
      self.grid = [*chain(*map(self.__ravel_group, self.grid))]
      
    grid_s = lambda s: product(*map(s.get, self.grid_fields))
    grid_iter = chain(*map(grid_s, self.grid))
    grid_iter = [dict(zip(self.grid_fields, i)) for i in grid_iter]
    
    return grid_iter
  
  def __getitem__(self, idx):
    if isinstance(idx, int):
      return ConfigDict({**self.static_configs, **self.grid_iter[idx]})
    elif isinstance(idx, slice):
      new_ci = copy.deepcopy(self)
      new_ci.grid_iter = new_ci.grid_iter[idx]
      return new_ci
      
  def __len__(self):
    return len(self.grid_iter)
    
  
  

@dataclass
class ExperimentLog:
  static_configs: dict
  grid_fields: list
  logs_file: str
  
  metric_fields: Optional[list] = None
  df: Optional[pd.DataFrame]=None
  auto_update_tsv: bool = False
  
  __sep: ClassVar[str] = '-'*45 + '\n'
  
  def __post_init__(self):
    if self.df is None:
      assert self.metric_fields is not None, 'Specify the metric fields of the experiment.'
      assert not (f:=set(self.grid_fields) & set(self.metric_fields)), f'Overlapping field names {f} in grid_fields and metric_fields. Remove one of them.'
      self.df = pd.DataFrame(columns=self.grid_fields+self.metric_fields).set_index(self.grid_fields)
    else:
      self.metric_fields = list(self.df)
  
  # Constructors.
  # -----------------------------------------------------------------------------  
  @classmethod
  def from_exp_config(cls, exp_config, logs_file: str, metric_fields: Optional[list]=None, auto_update_tsv: bool=False):
    return cls(*(exp_config[k] for k in ['static_configs', 'grid_fields']), logs_file=logs_file,
               metric_fields=metric_fields, auto_update_tsv = auto_update_tsv)

  @classmethod
  def from_tsv(cls, logs_file: str, parse_str=True, auto_update_tsv: bool=False):
    '''open tsv with yaml header'''
    return cls(**cls.parse_tsv(logs_file, parse_str=parse_str), logs_file=logs_file, auto_update_tsv=auto_update_tsv)
  
  
  # tsv handlers.
  # -----------------------------------------------------------------------------
  @classmethod
  def parse_tsv(cls, log_file: str, parse_str=True):
    '''parses tsv file into usable datas'''
    assert os.path.exists(log_file), f'File path "{log_file}" does not exists.'

    with open(log_file, 'r') as fd:
      # process yaml config header
      def header():
        next(fd)
        header = ''
        for s in fd:
          if s==cls.__sep: break
          header += s
        return header
    
      # get workload data from yaml header
      static_configs = yaml.safe_load(header())

      # get dataframe from csv body
      csv_str = fd.read()
      
      csv_col, csv_idx, *csv_body = csv_str.split('\n')
      col = csv_col.strip().split('\t')
      idx = csv_idx.strip().split('\t')
      csv_head = '\t'.join(idx+col)
      csv_str = '\n'.join([csv_head, *csv_body])
      
      df = pd.read_csv(io.StringIO(csv_str), sep='\t')
      df = df.drop(['id'], axis=1)
      
      if parse_str:
          df = df.applymap(str2value)
      
      # set grid_fields to multiindex
      df = df.set_index(idx[1:])
      
    return {'static_configs': static_configs,
            'grid_fields': idx[1:],
            'metric_fields': list(df),
            'df': df}
  

  def load_tsv(self, logs_file, parse_str=True):
    '''load tsv with yaml header'''
    if logs_file is not None:
      self.logs_file=logs_file
      
    for k, v in self.parse_tsv(self.logs_file, parse_str=parse_str).items():
      self.__dict__[k] = v
  

  def to_tsv(self, logs_file=None):
    logs_file = self.logs_file if logs_file==None else logs_file
    
    logs_path, _ = os.path.split(logs_file)
    if not os.path.exists(logs_path):
        os.makedirs(logs_path) 
    
    with open(logs_file, 'w') as fd:
      # write static_configs
      fd.write('[Static Configs]\n')
      yaml.dump(self.static_configs, fd)
      fd.write(self.__sep)

      # write table of results
      df = self.df.reset_index()
      df['id'] = [*range(len(df))]
      df = df.set_index(['id', *self.grid_fields])
      csv_str = df.to_csv(sep='\t')
      
      csv_head, *csv_body = csv_str.split('\n')
      csv_head = csv_head.split('\t')
      col = '\t'.join([' '*len(i) if i in df.index.names else i for i in csv_head])
      idx = '\t'.join([i if i in df.index.names else ' '*len(i) for i in csv_head])
      csv_str = '\n'.join([col, idx, *csv_body])
      
      fd.write(csv_str)
      
  
  def update_tsv(func, mode='rw'):
    '''Decorator for read/write tsv before/after given function call'''
    def wrapped(self, *args, **kwargs):
      if self.auto_update_tsv and 'r' in mode: 
        self.load_tsv(self.logs_file)
      ret = func(self, *args, **kwargs)
      if self.auto_update_tsv and 'w' in mode: self.to_tsv()
      return ret
    return wrapped

  
  # Add results.
  # -----------------------------------------------------------------------------
  
  @partial(update_tsv, mode='r')
  def add_result(self, configs, **metrics):
    '''Add experiment run result to dataframe'''
    if configs in self:
      cur_gridval = list2tuple([configs[k] for k in self.grid_fields])
      self.df = self.df.drop(cur_gridval)
    
    configs = {k:list2tuple(configs[k]) for k in self.grid_fields}
    metrics = {k:metrics.get(k) for k in self.metric_fields}
    result_dict = {k:[v] for k, v in {**configs, **metrics}.items()}
    result_df = pd.DataFrame(result_dict).set_index(self.grid_fields)
    self.df = pd.concat([self.df, result_df])[self.metric_fields]
    
    # Dataframe.loc based code
    # problem: pandas unable to distinguish between 
    #              ``.loc[(col_idx, row_idx)]`` and ``.loc[(multi_idx_1, multi_idx_2)]``
    #          for 2-level multiindex dataframe
    """
    cur_gridval = list2tuple([configs[k] for k in self.grid_fields])
    
    df_row = [metrics.get(k) for k in self.field_order]
      
    # Write over metric results if there is a config saved
    if configs in self:
      self.df = self.df.drop(cur_gridval)
    
    self.df.loc[cur_gridval] = df_row
    """
    
  @staticmethod
  def __add_column(df, new_column_name, fn, *fn_arg_fields):
    '''Add new column field computed from existing fields in self.df'''
    def mapper(*args):
      if all(isinstance(i, (int, float, str, tuple, list)) for i in args):
        return fn(*args)
      return None
    df[new_column_name] = df.apply(lambda df: mapper(*[df[c] for c in fn_arg_fields]), axis=1)
    return df

  def add_computed_metric(self, new_metric_name, fn, *fn_arg_fields):
    '''Add new metric computed from existing metrics in self.df'''
    self.df = self.__add_column(self.df, new_metric_name, fn, *fn_arg_fields)
    self.metric_fields.append(new_metric_name)
    
  def add_derived_index(self, new_index_name, fn, *fn_arg_fields):
    '''Add new index field computed from existing fields in self.df'''
    df = self.df.reset_index(self.grid_fields)
    df = self.__add_column(df, new_index_name, fn, *fn_arg_fields)
    self.grid_fields.append(new_index_name)
    self.df = df.set_index(self.grid_fields)
    
  def remove_metric(self, *metric_names):
    self.df = self.df.drop(columns=[*metric_names])
    self.metric_fields = [m for m in self.grid_fields if m not in metric_names]
    
  def remove_index(self, *field_names):
    self.df = self.df.reset_index([*field_names], drop=True)
    self.grid_fields = [f for f in self.grid_fields if f not in field_names]

  # Merge ExperimentLogs.
  # -----------------------------------------------------------------------------
  def __merge_one(self, other, same=True):
    '''
    Merge two logs into self.
    - The order of grid_fields follows self.
    - Difference between static_configs are moved to grid_fields.
    - If grid_fields are different between self & other
       - If it exists in static_configs, they are moved to grid_fields.
       - else it is filled with np.nan
    '''
    if same:
      assert self==other, 'Different experiments cannot be merged by default.'

    # find different fixed configs
    def same_diff(dictl, dictr):
      keys = set(dictl.keys()) & set(dictr.keys())
      same, diff = dict(), []
      for k in keys:
        if dictl[k]==dictr[k]: same[k]=dictl[k]
        else: diff.append(k)
      return same, diff
    
    new_sttc, diff_sttc = same_diff(self.static_configs, other.static_configs)

    # find new grid_fields
    new_to_self_sf = [sf for sf in other.grid_fields if sf not in self.grid_fields] + diff_sttc
    new_to_othr_sf = [sf for sf in self.grid_fields if sf not in other.grid_fields] + diff_sttc

    # fill in new grid_fields in each df from static_configs and configs
    # change list configs to tuple for hashablilty
    for sf in new_to_self_sf:
      self.df[sf] = [list2tuple(self.static_configs.get(sf, np.nan))]*len(self)

    for sf in new_to_othr_sf:
      other.df[sf] = [list2tuple(other.static_configs.get(sf, np.nan))]*len(other)

    self.static_configs = new_sttc
    self.grid_fields += new_to_self_sf
    
    self.df, other.df = (obj.df.reset_index() for obj in (self, other))
    self.df = pd.concat([self.df, other.df]) \
                .set_index(self.grid_fields)[self.metric_fields]
    return self

  def merge(self, *others, same=True):
    '''Merge multiple logs into self'''
    for other in others:
      self.__merge_one(other, same=same)

  @staticmethod
  def merge_tsv(*names, logs_path, save_path=None, same=True):
    if save_path is None:
      save_path = os.path.join(logs_path, 'log_merged.tsv')
    base, *logs = [ExperimentLog.from_tsv(os.path.join(logs_path, n+'.tsv'), parse_str=False) for n in names]
    base.merge(*logs, same=same)
    base.to_tsv(save_path)

  @staticmethod
  def merge_folder(logs_path, save_path=None, same=True):
    """change later if we start saving tsvs to other directories"""
    os.chdir(logs_path)
    logs = [f[:-4] for f in glob.glob("*.tsv")]
    ExperimentLog.merge_tsv(*logs, logs_path=logs_path, save_path=save_path, same=same)
    
  
  # Utilities.
  # -----------------------------------------------------------------------------

  def __cfg_match_row(self, config):
    grid_filt = reduce(lambda l, r: l & r, 
                       (self.df.index.get_level_values(k)==(str(config[k]) if isinstance(config[k], list) else config[k]) 
                        for k in self.grid_fields))
    return self.df[grid_filt]
  
  
  @partial(update_tsv, mode='r')
  def isin(self, config):
    '''Check if specific experiment config was already executed in log.'''
    if self.df.empty: return False

    cfg_same_in_static = all([config[k]==v for k, v in self.static_configs.items() if k in config])
    cfg_matched_df = self.__cfg_match_row(config)
    
    return cfg_same_in_static and not cfg_matched_df.empty


  def get_metric(self, config):
    '''Search matching log with given config dict and return metric_dict, info_dict'''
    assert config in self, 'config should be in self when using get_metric_dict.'
    
    cfg_matched_df = self.__cfg_match_row(config)
    metric_dict = {k:(v.iloc[0] if not (v:=cfg_matched_df[k]).empty else None) for k in self.metric_fields}
    return metric_dict

  def is_same_exp(self, other):
    '''Check if both logs have same config fields.'''
    fields = lambda log: set(log.static_configs.keys()) | set(log.grid_fields)
    return fields(self)==fields(other)
    
    
  def melt_and_explode_metric(self, df=None, step=None):
    if df is None: 
      df = self.df
    mov_to_index = lambda *fields: df.reset_index().set_index([*df.index.names, *fields])
    
    # melt
    df = df.melt(value_vars=list(df), var_name='metric', value_name='metric_value', ignore_index=False)
    df = mov_to_index('metric')
    
    # Create step field and explode
    pseudo_len = lambda x: len(x) if isinstance(x, list) else 1
    
    df['total_steps'] = df['metric_value'].map(pseudo_len)
    
    if step is None:
        df['step'] = df['metric_value'].map(lambda x: range(1, pseudo_len(x)+1))
        df = df.explode('step')  # explode metric list so each step gets its own row
    else:
        df['step'] = df['metric_value'].map(lambda x: step + (pseudo_len(x) if step<0 else 0))
    
    df['metric_value'] = df.apply(lambda df: df['metric_value'][df.step-1] if isinstance(df['metric_value'], list) else df['metric_value'], axis=1) # list[epoch] for all fields
    
    df = mov_to_index('step', 'total_steps')
    
    # delete string and NaN valued rows
    df = df[pd.to_numeric(df['metric_value'], errors='coerce').notnull()]\
              .dropna()\
              .astype('float')
    
    return df

    
  def __contains__(self, config):
    return self.isin(config)

  def __eq__(self, other):
    return self.is_same_exp(other)

  def __len__(self):
    return len(self.df)

  def __str__(self):
    return '[Static Configs]\n' + \
           '\n'.join([f'{k}: {v}' for k,v in self.static_configs.items()]) + '\n' + \
           self.__sep + \
           str(self.df)




class Experiment:
  '''
  Executes experiments according to experiment configs
  
  Following is supported
  - Provides 2 methods parallel friedly experiments scheduling (can choose with bash arguments).
    - (plan splitting) Splits experiment plans evenly.
    - (current run checking) Save configs of currently running experiments to tsv so other running code can know.
  - Saves experiment logs, automatically resumes experiment using saved log.
  '''
  __RUNNING: ClassVar[str] = 'R'
  __FAILED: ClassVar[str] = 'F'
  __COMPLETED: ClassVar[str] = 'C'
  
  infos: ClassVar[list] = ['datetime', 'duration', 'commit_hash', 'status']
  
  def __init__(self, 
               exp_folder_path: str,
               exp_function: ExpFunc,
               exp_metrics: Optional[list] = None,
               total_splits: Union[int, str] = 1, 
               curr_split: Union[int, str] = 0,
               auto_update_tsv: bool = False,
               configs_save: bool = False,
               checkpoint: bool = False
    ):
    
    if checkpoint:
      assert auto_update_tsv, "argument 'auto_update_tsv' should be set to True when checkpointing."
    
    self.exp_func = exp_function

    self.exp_bs = total_splits
    self.exp_bi = curr_split
    self.configs_save = configs_save
    self.checkpoint = checkpoint
    
    cfg_file, tsv_file, _ = self.get_paths(exp_folder_path)
    self.configs = ConfigIter(cfg_file)
    self.__process_split()

    if isinstance(self.exp_bs, int) and self.exp_bs>1 or isinstance(self.exp_bs, str):
      tsv_file = os.path.join(exp_folder_path, 'log_splits', f'split_{self.exp_bi}.tsv') # for saving seperate log for each split in plan slitting mode.
    
    exp_metrics = self.infos + exp_metrics
    self.log = self.__get_log(tsv_file, exp_metrics, auto_update_tsv)
    
    
  def __process_split(self):
    
    assert self.exp_bs.isdigit() or (self.exp_bs in self.configs.grid_fields), \
        f'Enter valid splits (int | Literal{self.configs.grid_fields}).'
    # if total exp split is given as integer : uniformly split
    if self.exp_bs.isdigit():
      self.exp_bs, self.exp_bi = map(int, [self.exp_bs, self.exp_bi])
      assert self.exp_bs > 0, 'Total number of experiment splits should be larger than 0'
      assert self.exp_bs > self.exp_bi, 'Experiment split index should be smaller than the total number of experiment splits'
      if self.exp_bs>1:
        self.configs.filter_iter(lambda i, _: i%self.exp_bs==self.exp_bi)
        
    # else split across certain study field
    elif self.exp_bs in self.configs.grid_fields:
      
      self.exp_bi = [*map(str2value, self.exp_bi.split())]
      self.configs.filter_iter(lambda _, d: d[self.exp_bs] in self.exp_bi)
      
      
      
  def __get_log(self, logs_file, metric_fields=None, auto_update_tsv=False):
    # Configure experiment log
    if os.path.exists(logs_file): # Check if there already is a file
      log = ExperimentLog.from_tsv(logs_file, auto_update_tsv=auto_update_tsv) # resumes automatically
    else: # Create new log
      logs_path, _ = os.path.split(logs_file)
      if not os.path.exists(logs_path):
        os.makedirs(logs_path)
      log = ExperimentLog.from_exp_config(self.configs.__dict__, logs_file,
                                          metric_fields=metric_fields, auto_update_tsv=auto_update_tsv)
      log.to_tsv()
    return log
  
  
  @staticmethod
  def get_paths(exp_folder):
    cfg_file = os.path.join(exp_folder, 'exp_config.yaml')
    tsv_file = os.path.join(exp_folder, 'log.tsv')
    fig_dir = os.path.join(exp_folder, 'figure')
    return cfg_file, tsv_file, fig_dir
  
  def get_log_checkpoint(self, config):
    metric_dict = self.log.get_metric(config)
    info_dict = {k:v for k in self.infos if (k in metric_dict and pd.notna(v:=metric_dict.pop(k)))}
    metric_dict = {k:v for k, v in metric_dict.items() if pd.notna(v)}
    return metric_dict, info_dict
    
  def update_log(self, config, **metric_dict):
    self.log.add_result(config, **metric_dict, 
                        datetime=str(datetime.now()), 
                        status=self.__RUNNING)
    self.log.to_tsv()
    
  def run(self):
    
    try:
      commit_hash = Repo.init().head.commit.hexsha
    except:
      commit_hash = None
      logging.info('No git exist in current directory.')
    
    # current experiment count
    if isinstance(self.exp_bs, int):
      logging.info(f'Experiment : {self.configs.name} (split : {self.exp_bi+1}/{self.exp_bs})')
    elif isinstance(self.exp_bs, str):
      logging.info(f'Experiment : {self.configs.name} (split : {self.exp_bi}/{self.configs.grid_dict[self.exp_bs]})')
    
    # run experiment plans 
    for i, config in enumerate(self.configs):

      if config in self.log:
        metric_dict, info_dict = self.get_log_checkpoint(config)
        if info_dict.get('status') != self.__FAILED:
          continue # skip already executed runs
      
      # if config not in self.log or status==self.__FAILED
      if self.configs_save:
        if config not in self.log: metric_dict = {}
        self.log.add_result(config, **metric_dict,
                            status=self.__RUNNING)
        self.log.to_tsv()

      logging.info('###################################')
      logging.info(f'   Experiment count : {i+1}/{len(self.configs)}')
      logging.info('###################################') 


      start_t = datetime.now()
      try:
        if self.checkpoint:
          metric_dict = self.exp_func(config, self)
        else:
          metric_dict = self.exp_func(config)
        status = self.__COMPLETED
        
      except:
        metric_dict = self.get_log_checkpoint(config)[0] if config in self.log else {}
        status = self.__FAILED
        raise
      
      finally:
        end_t = datetime.now()
        prev_duration = datetime.timedelta(info_dict['duration'] if (config in self.log and 'duration' in info_dict) else 0)
        self.log.add_result(config, **metric_dict,
                            datetime=end_t, 
                            duration=prev_duration+(end_t-start_t),
                            commit_hash=commit_hash,
                            status=status)
        self.log.to_tsv()
        
        logging.info("Saved experiment data to log")
      
      
      
  @staticmethod
  def resplit_logs(exp_folder_path: str, target_split: int=1, save_backup: bool=True):
    """Resplit splitted logs into ``target_split`` number of splits."""
    assert target_split > 0, 'Target split should be larger than 0'
    
    cfg_file, logs_file, _ = Experiment.get_paths(exp_folder_path)
    logs_folder = os.path.join(exp_folder_path, 'log_splits')
    
    # merge original log_splits
    if os.path.exists(logs_folder): # if log is splitted
      os.chdir(logs_folder)
      base, *logs = [ExperimentLog.from_tsv(os.path.join(logs_folder, sp_n), parse_str=False) for sp_n in glob.glob("*.tsv")]
      base.merge(*logs)
      shutil.rmtree(logs_folder)
    elif os.path.exists(logs_file): # if only single log file exists 
      base = ExperimentLog.from_tsv(os.path.join(logs_file), parse_str=False)
      shutil.rmtree(logs_file)
    
    # save backup
    if save_backup:
      base.to_tsv(os.path.join(exp_folder_path, 'logs_backup.tsv'))
    
    # resplit merged logs based on target_split
    if target_split==1:
      base.to_tsv(logs_file)
    
    elif target_split>1:
      # get configs
      configs = ConfigIter(cfg_file)
      
      for n in range(target_split):
        # empty log
        lgs = ExperimentLog.from_exp_config(configs.__dict__, 
                                            os.path.join(logs_folder, f'split_{n}.tsv',),
                                            base.metric_fields)
        
        # resplitting nth split
        cfgs_temp = copy.deepcopy(configs)
        cfgs_temp.filter_iter(lambda i, _: i%target_split==n)
        for cfg in track(cfgs_temp, description=f'split: {n}/{target_split}'):
          if cfg in base:
            metric_dict = base.get_metric(cfg)
            lgs.add_result(cfg, **metric_dict)
          
        lgs.to_tsv()




import os, shutil
from ast import literal_eval

from rich.table import Table

def create_dir(dir):
  if os.path.exists(dir):
    for f in os.listdir(dir):
      if os.path.isdir(os.path.join(dir, f)):
        shutil.rmtree(os.path.join(dir, f))
      else:
        os.remove(os.path.join(dir, f))
  else:
    os.makedirs(dir)
    

def df2richtable(df):
  table = Table(title='Metric Summary Table')
  df = df.reset_index()
  
  table.add_column('id')
  for f in list(df):       
    table.add_column(f)
  
  for row in df.itertuples(name=None):
    table.add_row(*(str(i) for i in row))
    
  return table


def list2tuple(l):
  if isinstance(l, list):
    return tuple(map(list2tuple, l))
  if isinstance(l, dict):
    return {k: list2tuple(v) for k, v in l.items()}
  return l
    

def str2value(value_str):
    """Casts string back to standard python types"""
    if not isinstance(value_str, str): return value_str
    value_str = value_str.replace('inf', '2e+308')
    try:
      return literal_eval(value_str)
    except:
      return value_str


from typing import Literal

import numpy as np
import pandas as pd
from matplotlib.axes import Axes


def ax_draw_curve(ax: Axes,
                  df: pd.DataFrame,
                  label: str,
                  annotate = True,
                  annotate_field = [],
                  std_plot: Literal['none','fill','bar'] = 'fill',
                  unif_xticks = False,
                  color = 'orange', 
                  linewidth = 4, 
                  marker = 'D', 
                  markersize = 10, 
                  markevery = 20,
                  **_
    ) -> Axes:
    """
    Draws curve of y_field over arbitrary x_field setted as index of the dataframe.
    If there is column 'y_field_sdv' is in dataframe, draws in errorbar or fill_between depending on ``sdv_bar_plot``
    """
    assert std_plot in {'bar', 'fill', 'none'}, 'std_plot should be one of {"bar","fill","none"}'
    y_field = list(df)[0]
    
    x_values, metric_values = map(np.array, zip(*dict(df[y_field]).items()))
    assert not isinstance(metric_values[0], pd.Series), 'y_field should have only 1 values for each index.'
    
    if len(x_values)>100:
        marker = None
        if unif_xticks: ax.locator_params(tight=True, nbins=5)
    elif len(x_values)>20:
        markevery = 20
        if unif_xticks: ax.locator_params(tight=True, nbins=5)
    else:
        markevery = 1
    
    tick_values = x_values
    
    if len(tick_values)==1:
        ax.axhline(metric_values, linewidth=linewidth, color=color, label=label)
        if f'{y_field}_std' in df:
            metric_std = float(df[f'{y_field}_std'])
            ax.axhspan(metric_values[0] + metric_std, metric_values[0] - metric_std, alpha=0.3, color=color)

    else:
        if unif_xticks:
            tick_values = np.arange(len(x_values))
            ax.set_xticks(tick_values, x_values, fontsize=10, rotation=45)
            
        ax.plot(tick_values, metric_values, label=label, color=color, linewidth=linewidth, 
                marker=marker, markersize=markersize, markevery=markevery)
        if f'{y_field}_std' in df:
            x_values, metric_std = map(np.array, zip(*dict(df[f'{y_field}_std']).items()))
            
            if std_plot=='bar':
                ax.errorbar(tick_values, metric_values, yerr=metric_std, color=color, elinewidth=3)
            elif std_plot=='fill':
                ax.fill_between(tick_values, metric_values + metric_std, metric_values - metric_std, alpha=0.3, color=color)
                
        if annotate:
            assert not (f:=set(annotate_field) - (a:=set(df) - {'total_epochs', 'epoch', y_field, f'{y_field}_std'})), f'Annotation field: {f} are not in dataframe field: {a}'
            annotate_field = set(annotate_field) & a
            abv = lambda s: ''.join([i[0] for i in s.split('_')] if '_' in s else \
                                      [s[0]] + [i for i in s[1:] if i not in 'aeiou']) if len(s)>3 else s
            abv_annot = [*map(abv, annotate_field)]
            for i, (x,y,t) in enumerate(zip(x_values, metric_values, tick_values)):
                if i%markevery: continue
                txt = '\n'.join([f'{y:.5f}', str(x)]+[f'{i}={df.loc[x][j]}' for i, j in zip(abv_annot, annotate_field)])
                ax.annotate(txt, (t,y), textcoords="offset points", xytext=(0,10), ha='center')
    
    ax.tick_params(axis='both', which='major', labelsize=17, direction='in', length=5)

    return ax

def ax_draw_best_stared_curve(ax: Axes,
                  df: pd.DataFrame,
                  label: str,
                  annotate = True,
                  annotate_field = [],
                  std_plot: Literal['none','fill','bar'] = 'fill',
                  best_at_max=True,
                  unif_xticks = False,
                  color = 'orange', 
                  linewidth = 4, 
                  marker = 'D', 
                  markersize = 10, 
                  markevery = 20,
                  **_
    ) -> Axes:
    """
    Draws curve of y_field over arbitrary x_field setted as index of the dataframe.
    If there is column 'y_field_sdv' is in dataframe, draws in errorbar or fill_between depending on ``sdv_bar_plot``
    """
    assert std_plot in {'bar', 'fill', 'none'}, 'std_plot should be one of {"bar","fill","none"}'
    y_field = list(df)[0]
    
    x_values, metric_values = map(np.array, zip(*dict(df[y_field]).items()))
    assert not isinstance(metric_values[0], pd.Series), 'y_field should have only 1 values for each index.'
    
    if len(x_values)>100:
        marker = None
        if unif_xticks: ax.locator_params(tight=True, nbins=5)
    elif len(x_values)>20:
        markevery = 20
        if unif_xticks: ax.locator_params(tight=True, nbins=5)
    else:
        markevery = 1
    
    tick_values = x_values
    
    if len(tick_values)==1:
        ax.axhline(metric_values, linewidth=linewidth, color=color, label=label)
        if f'{y_field}_std' in df:
            metric_std = float(df[f'{y_field}_std'])
            ax.axhspan(metric_values[0] + metric_std, metric_values[0] - metric_std, alpha=0.3, color=color)

    else:
        if unif_xticks:
            tick_values = np.arange(len(x_values))
            ax.set_xticks(tick_values, x_values, fontsize=10, rotation=45)
            
        ax.plot(tick_values, metric_values, color=color, linewidth=linewidth)
        if f'{y_field}_std' in df:
            x_values, metric_std = map(np.array, zip(*dict(df[f'{y_field}_std']).items()))
            
            if std_plot=='bar':
                ax.errorbar(tick_values, metric_values, yerr=metric_std, color=color, elinewidth=3)
            elif std_plot=='fill':
                ax.fill_between(tick_values, metric_values + metric_std, metric_values - metric_std, alpha=0.3, color=color)
                
        best_idx = list(metric_values).index((max if best_at_max else min)(metric_values))
        for i, (_,y,t) in enumerate(zip(x_values, metric_values, tick_values)):
            if i%markevery: continue
            if i==0:
                ax.plot(tick_values[i], metric_values[i], label=label, color=color, linewidth=linewidth,
                        marker=marker, markersize=markersize, markevery=markevery)
            elif i==best_idx:
                ax.plot(tick_values[i], metric_values[i], color='green', marker='*', markersize=markersize+10)
            else:
                ax.plot(tick_values[i], metric_values[i], color=color, 
                        marker=marker, markersize=markersize, markevery=markevery)
                
                
        if annotate:
            assert not (f:=set(annotate_field) - (a:=set(df) - {'total_epochs', 'epoch', y_field, f'{y_field}_std'})), f'Annotation field: {f} are not in dataframe field: {a}'
            annotate_field = set(annotate_field) & a
            abv = lambda s: ''.join([i[0] for i in s.split('_')] if '_' in s else \
                                      [s[0]] + [i for i in s[1:] if i not in 'aeiou']) if len(s)>3 else s
            abv_annot = [*map(abv, annotate_field)]
            for i, (x,y,t) in enumerate(zip(x_values, metric_values, tick_values)):
                if i%markevery: continue
                txt = '\n'.join([f'{y:.5f}', str(x)]+[f'{i}={df.loc[x][j]}' for i, j in zip(abv_annot, annotate_field)])
                ax.annotate(txt, (t,y), textcoords="offset points", xytext=(0,10), ha='center')
    
    ax.tick_params(axis='both', which='major', labelsize=17, direction='in', length=5)

    return ax


def ax_draw_bar(ax: Axes,
                df: pd.DataFrame,
                label: str,
                annotate = True,
                annotate_field = [],
                std_plot = True,
                unif_xticks = False,
                color = 'orange',
                **_
    ) -> Axes:
    """
    Draws bar graph of y_field over arbitrary x_field setted as index of the dataframe.
    If there is column 'y_field_sdv' is in dataframe, draws in errorbar or fill_between depending on ``sdv_bar_plot``
    """
    y_field = list(df)[0]
    
    x_values, metric_values = map(np.array, zip(*dict(df[y_field]).items()))
    assert not isinstance(metric_values[0], pd.Series), 'y_field should have only 1 values for each index.'
    
    tick_values = np.arange(len(x_values))
    ax.set_xticks(tick_values, x_values, fontsize=10, rotation=45)

    if std_plot and f'{y_field}_std' in df:
        x_values, metric_std = map(np.array, zip(*dict(df[f'{y_field}_std']).items()))
        ax.bar(tick_values, metric_values, yerr=metric_std, label=label, color=color)
    else:
        ax.bar(tick_values, metric_values, label=label, color=color)
        
    if annotate:
        assert not (f:=set(annotate_field) - (a:=set(df) - {'total_epochs', 'epoch', y_field, f'{y_field}_std'})), f'Annotation field: {f} are not in dataframe field: {a}'
        annotate_field = set(annotate_field) & a
        abv = lambda s: ''.join([i[0] for i in s.split('_')] if '_' in s else \
                                [s[0]] + [i for i in s[1:] if i not in 'aeiou']) if len(s)>3 else s
        abv_annot = [*map(abv, annotate_field)]
        for x,y,t in zip(x_values, metric_values, tick_values):
            txt = '\n'.join([f'{y:.5f}']+['' if unif_xticks else str(x)]+[f'{i}={df.loc[x][j]}' for i, j in zip(abv_annot, annotate_field)])
            ax.annotate(txt, (t,y), textcoords="offset points", xytext=(0,10), ha='center')
    
    ax.tick_params(axis='both', which='major', labelsize=17, direction='in', length=5)

    return ax


def ax_draw_heatmap(ax: Axes,
                    df: pd.DataFrame,
                    cmap = 'magma', 
                    annotate=True,
                    annotate_field=[],
                    **_
        ) -> Axes:
    """
    Draws heatmap of y_field over two arbitrary x_fields setted as multi-index of the dataframe.
    """
    
    grid_df = df.drop(columns=[list(df)[i] for i in range(1, len(df.columns))])
    y_field = list(df)[0]
    
    x_fields = grid_df.index.names
    *x_values, = map(lambda l: sorted(set(grid_df.index.get_level_values(l))), x_fields)
    grid_df = grid_df.reset_index()\
           .pivot(index=x_fields[1], columns=x_fields[0])
    
    ax.pcolor(grid_df, cmap=cmap, edgecolors='w')
    
    ax.set_xticks(np.arange(0.5, len(x_values[0]), 1), x_values[0], fontsize=10, rotation=45)
    ax.set_yticks(np.arange(0.5, len(x_values[1]), 1), x_values[1], fontsize=10)
        
    if annotate:
        assert not (f:=set(annotate_field) - (a:=set(df) - {'total_epochs', 'epoch', y_field, f'{y_field}_std'})), f'Annotation field: {f} are not in dataframe field: {a}'
        annotate_field = set(annotate_field) & a
        abv = lambda s: ''.join([i[0] for i in s.split('_')] if '_' in s else \
                                [s[0]] + [i for i in s[1:] if i not in 'aeiou']) if len(s)>3 else s
        abv_annot = [*map(abv, annotate_field)]
        
        for i, (mtc, x) in enumerate([*grid_df]):
            for j, y in enumerate([*grid_df.index.get_level_values(0)]):
                txt = '\n'.join([f'{grid_df.loc[y, (mtc, x)]:.5f}']+[f'{i}={df.loc[(x, y), j]}' for i, j in zip(abv_annot, annotate_field)])
                ax.text(i+0.5, j+0.5, txt, c='dimgrey', ha='center', va='center', weight='bold')
    
    # ax.tick_params(axis='both', which='major', labelsize=17, direction='in', length=5)
    return ax



import os, shutil
from operator import attrgetter

from matplotlib.axes import Axes

def create_dir(dir):
  if os.path.exists(dir):
    for f in os.listdir(dir):
      if os.path.isdir(os.path.join(dir, f)):
        shutil.rmtree(os.path.join(dir, f))
      else:
        os.remove(os.path.join(dir, f))
  else:
    os.makedirs(dir)

def save_figure(fig, save_dir, save_name, png=False):
    fig.tight_layout()
    if not os.path.exists(save_dir):
        create_dir(save_dir)
    save_path = os.path.join(save_dir, save_name + '.pdf')
    fig.savefig(save_path, format='pdf')

    if png:
        png_save_dir = save_path.replace('/pdf', '/png')
        if not os.path.exists(png_save_dir):
            create_dir(png_save_dir)
        fig.savefig(os.path.join(png_save_dir, save_name + '.png'), bbox_inches='tight', format='png')
      
def merge_dict(base: dict, other: dict):
    """Merge plot_config dict (priority: ``base``)"""
    for k in (set(base) & set(other)):
        if isinstance(base[k], list):
            if base[k] and isinstance(base[k][-1], dict):
              base[k] = base[k][:-1] + other[k][:-1] \
                        + [merge_dict(base[k][-1], other[k][-1])]
        elif isinstance(base[k], dict):
            base[k] = merge_dict(base[k], other[k])
    for k in (set(other) - set(base)):
        base[k] = other[k]
    return base

default_style = {
    'annotate': False,
    'std_plot': 'fill',
    'ax_style':{
        'frame_width': 2.5,
        'fig_size': 7,
        'legend': [{'fontsize': 20}],
        'grid': [True, {'linestyle': '--'}],
        'tick_params':[{'axis': 'both',
                        'which': 'major',
                        'labelsize': 25,
                        'direction': 'in',
                        'length': 5}]
    },
    'line_style': {
        'linewidth': 4,
        'marker': 'D',
        'markersize': 10,
        'markevery': 1,
    }
}

def ax_styler(ax: Axes, **style_dict):
  if (n:='fig_size') in style_dict:
    dim = style_dict.pop(n)
    if isinstance(dim, int):
      w = h = dim
    else:
      w, h = dim
    ax.figure.set_figwidth(w)
    ax.figure.set_figheight(h)
  
  if (n:='frame_width') in style_dict:
    fw = style_dict.pop(n)
    for axis in ['top', 'bottom', 'left', 'right']:
      ax.spines[axis].set_linewidth(fw)
  
  non_set = ['tick_params', 'legend', 'grid']
  for name, (*arg_pos, arg_kw) in style_dict.items():
    attr_name = name if name in non_set else f'set_{name}'
    attrgetter(attr_name)(ax)(*arg_pos, **arg_kw)
    
  


